// -------------------------------------------------------------------------------- 
//!  @file  ngplcd.c                                                                
//!  NEOGEO POCKET LCD データ変換ソースファイル                                     
//                                                                                  
//   Copyright (C) 2023 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include <pico/stdlib.h>
#include "config.h"
#include "vga.h"
#include "ngplcd.h"
#include "ngplcd.pio.h"

#if     (NGPLCD_CFG_PIO == 0)
#define PIO							pio0							//!< PIO オブジェクト
#define PIO_IRQ						PIO0_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO0_RX0					//!< PIO データリクエスト
#else
#define PIO							pio1							//!< PIO オブジェクト
#define PIO_IRQ						PIO1_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO1_RX0					//!< PIO データリクエスト
#endif/*(NGPLCD_CFG_PIO == 0)*/

#if     (NGPLCD_CFG_DMA == 0)
#define DMA_IRQ						DMA_IRQ_0						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints0							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq0_enabled	//!< DMA 有効処理
#else
#define DMA_IRQ						DMA_IRQ_1						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints1							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq1_enabled	//!< DMA 有効処理
#endif//(NGPLCD_CFG_DMA == 0)


static dma_channel_config			m_dmaConfigPixel ;
static uint32_t						m_ngpData[NGPLCD_HEIGHT][7][NGPLCD_WIDTH/32] ;
static uint8_t						m_color ;

static uint16_t						m_convColor[][8]={
										{		// モノクロ
											0x0000,0x0333,0x0444,0x0777,0x0999,0x0bbb,0x0ddd,0x0fff
										}, {	// 赤
											0x0000,0x0006,0x0009,0x011d,0x055f,0x099f,0x0ccf,0x0fff
										}, {	// 緑
											0x0000,0x0050,0x0080,0x00b0,0x03d3,0x07f7,0x0bfb,0x0fff
										}, {	// 青
											0x0000,0x0700,0x0b22,0x0e55,0x0f88,0x0faa,0x0fcc,0x0fff
										}, {	// セピア
											0x0000,0x0034,0x0157,0x0379,0x059b,0x08bd,0x0ade,0x0fff
										}, {	// モノクロ
											0x0000,0x0333,0x0444,0x0777,0x0999,0x0bbb,0x0ddd,0x0fff
										}, {	// モノクロ
											0x0000,0x0333,0x0444,0x0777,0x0999,0x0bbb,0x0ddd,0x0fff
										}
									} ;


// -------------------------------------------------------------------------------- 
//!  ８階調データから RGB データ変換                                                
//!                                                                                 
//!  @param [in]  p_srce   ８階調データ                                             
//!  @param [out] p_dest   RGB データ格納アドレス                                   
//!  @param [in]  p_color  色変換データ                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void RAMFUNC gradationToRGB(uint32_t *p_srce, uint16_t *p_dest, uint16_t *p_color)
{
	int			index ;
	uint32_t	data ;


	for(index=3; index>=0; index--,p_dest+=3) {
		data        = ((p_srce[0*5] >> index) & 0x11111111) +
		              ((p_srce[1*5] >> index) & 0x11111111) +
		              ((p_srce[2*5] >> index) & 0x11111111) +
		              ((p_srce[3*5] >> index) & 0x11111111) +
		              ((p_srce[4*5] >> index) & 0x11111111) +
		              ((p_srce[5*5] >> index) & 0x11111111) +
		              ((p_srce[6*5] >> index) & 0x11111111) ;

		p_dest[4*3*0+0] = p_dest[4*3*0+1] = p_dest[4*3*0+2] = p_color[(data>>(4*7))&7] ;
		p_dest[4*3*1+0] = p_dest[4*3*1+1] = p_dest[4*3*1+2] = p_color[(data>>(4*6))&7] ;
		p_dest[4*3*2+0] = p_dest[4*3*2+1] = p_dest[4*3*2+2] = p_color[(data>>(4*5))&7] ;
		p_dest[4*3*3+0] = p_dest[4*3*3+1] = p_dest[4*3*3+2] = p_color[(data>>(4*4))&7] ;
		p_dest[4*3*4+0] = p_dest[4*3*4+1] = p_dest[4*3*4+2] = p_color[(data>>(4*3))&7] ;
		p_dest[4*3*5+0] = p_dest[4*3*5+1] = p_dest[4*3*5+2] = p_color[(data>>(4*2))&7] ;
		p_dest[4*3*6+0] = p_dest[4*3*6+1] = p_dest[4*3*6+2] = p_color[(data>>(4*1))&7] ;
		p_dest[4*3*7+0] = p_dest[4*3*7+1] = p_dest[4*3*7+2] = p_color[(data>>(4*0))&7] ;
	}
}


// -------------------------------------------------------------------------------- 
//!  ８階調データを描画                                                             
//!                                                                                 
//!  @param [in]  p_srce   ８階調データ                                             
//!  @param [out] p_dest   RGB データ格納アドレス                                   
//!  @param [in]  p_color  色変換データ                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void RAMFUNC drawGradationData(uint32_t *p_srce, uint16_t *p_dest, uint16_t *p_color)
{
	gradationToRGB(&p_srce[0], &p_dest[32*3*0], &p_color[0]) ;
	gradationToRGB(&p_srce[1], &p_dest[32*3*1], &p_color[0]) ;
	gradationToRGB(&p_srce[2], &p_dest[32*3*2], &p_color[0]) ;
	gradationToRGB(&p_srce[3], &p_dest[32*3*3], &p_color[0]) ;
	gradationToRGB(&p_srce[4], &p_dest[32*3*4], &p_color[0]) ;
}


// -------------------------------------------------------------------------------- 
//!  DMA 割り込みサービスルーチン                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void RAMFUNC isr_dma( void )
{
	if( dma_hw->DMA_INTS & (1<<NGPLCD_CFG_DMACH_PIXEL) ) {
		dma_hw->DMA_INTS = (1<<NGPLCD_CFG_DMACH_PIXEL) ;
		dma_channel_configure(NGPLCD_CFG_DMACH_PIXEL, &m_dmaConfigPixel, &m_ngpData[0][0][0], &PIO->rxf[0], sizeof( m_ngpData )/sizeof( uint32_t ), true) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC ngplcd_init( void )
{
	pio_sm_config	config ;
	uint			offset ;


	gpio_init_mask( (1<<NGPLCD_CFG_PIN_DISP) |
	                (1<<NGPLCD_CFG_PIN_CL2)  |
	                (1<<NGPLCD_CFG_PIN_DI0) |
	                (1<<NGPLCD_CFG_PIN_DI1) |
	                (1<<NGPLCD_CFG_PIN_DI2) |
	                (1<<NGPLCD_CFG_PIN_DI3) ) ;

	gpio_pull_up( NGPLCD_CFG_PIN_DISP ) ;
	gpio_pull_up( NGPLCD_CFG_PIN_CL2 ) ;

	// データ読み取り設定
	offset = pio_add_program(PIO, &ngplcd_data_program) ;
	config = ngplcd_data_program_get_default_config( offset ) ;
	sm_config_set_in_pins(&config, NGPLCD_CFG_PIN_DI0) ;
	sm_config_set_in_shift(&config, false, true, 32) ;
	pio_sm_init(PIO, 0, offset, &config) ;

	// スクラッチレジスター Y 初期化
	pio_sm_put_blocking(PIO, 0, (((NGPLCD_WIDTH/4)*7*NGPLCD_HEIGHT-1)<<16)|((9203*2-50-1)<<0)) ;
	pio_sm_exec_wait_blocking(PIO, 0, pio_encode_pull(false, false)) ;
	pio_sm_exec_wait_blocking(PIO, 0, pio_encode_mov(pio_y, pio_osr)) ;

	// データ転送開始割り込み設定
	pio_interrupt_clear(PIO, 0) ;
	pio_set_irq0_source_enabled(PIO, pis_sm0_rx_fifo_not_empty, true) ;
	irq_set_exclusive_handler(DMA_IRQ, isr_dma) ;
	irq_clear( DMA_IRQ ) ;
	irq_set_enabled(DMA_IRQ, true) ;

	// ピクセル転送用 DMA
	dma_channel_claim( NGPLCD_CFG_DMACH_PIXEL ) ;
	m_dmaConfigPixel = dma_channel_get_default_config( NGPLCD_CFG_DMACH_PIXEL ) ;
	channel_config_set_read_increment(&m_dmaConfigPixel,  false) ;
	channel_config_set_write_increment(&m_dmaConfigPixel, true) ;
	channel_config_set_dreq(&m_dmaConfigPixel, PIO_DREQ) ;
	dma_channel_configure(NGPLCD_CFG_DMACH_PIXEL, &m_dmaConfigPixel, &m_ngpData[0][0][0], &PIO->rxf[0], sizeof( m_ngpData )/sizeof( uint32_t ), true) ;
	dma_enable(NGPLCD_CFG_DMACH_PIXEL, true) ;
}


// -------------------------------------------------------------------------------- 
//!  開始                                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void ngplcd_start( void )
{
	pio_sm_set_enabled(PIO, 0, true) ;
}


// -------------------------------------------------------------------------------- 
//!  色設定                                                                         
//!                                                                                 
//!  @param [in] color  色インデックス                                              
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
void RAMFUNC ngplcd_setColor( uint8_t color )
{
	m_color = color ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC ngplcd_waitVSync( void )
{
	while( !gpio_get( NGPLCD_CFG_PIN_DISP ) ) {
		;
	}

	while( gpio_get( NGPLCD_CFG_PIN_DISP ) ) {
		;
	}
}


// -------------------------------------------------------------------------------- 
//!  １行描画                                                                       
//!                                                                                 
//!  @param [out] p_buff  描画バッファアドレス                                      
//!  @param [in]  line    描画行                                                    
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
void RAMFUNC ngplcd_drawLine(uint16_t *p_buff, uint32_t line)
{
	int			index ;
	uint16_t	color ;


	g_vgaInfo.color = NGPVGA_CFG_FRAME ;

	// 背景塗り潰し開始
	p_buff[0] = g_vgaInfo.color ;
	p_buff[1] = g_vgaInfo.color ;
	channel_config_set_read_increment(&g_vgaInfo.dmaConfigData, false) ;
	dma_channel_configure(VGA_CFG_DMACH_DATA, &g_vgaInfo.dmaConfigData, &p_buff[0], &p_buff[0], (VGA_IMAGE_WIDTH*2)/4, true) ;

	line -= 4 ;
	if( line >= NGPLCD_HEIGHT ) {

		// 背景塗り潰し完了待ち
		dma_channel_wait_for_finish_blocking( VGA_CFG_DMACH_DATA ) ;

		if( ((int32_t )line==-1) || (line==NGPLCD_HEIGHT) ) {
			color = NGPVGA_CFG_BORDERUL ;
			if( line == NGPLCD_HEIGHT ) {
				color = NGPVGA_CFG_BORDERBR ;
			}
			for(index=80-3; index<80-3+(NGPLCD_WIDTH+2)*3; index+=3) {
				p_buff[index+0] = color ;
				p_buff[index+1] = color ;
				p_buff[index+2] = color ;
			}
		}
		return ;
	}

	// 背景塗り潰し完了待ち
	dma_channel_wait_for_finish_blocking( VGA_CFG_DMACH_DATA ) ;

	p_buff[80               -3] = NGPVGA_CFG_BORDERUL ;
	p_buff[80               -2] = NGPVGA_CFG_BORDERUL ;
	p_buff[80             -  1] = NGPVGA_CFG_BORDERUL ;
	p_buff[80+NGPLCD_WIDTH*3+0] = NGPVGA_CFG_BORDERBR ;
	p_buff[80+NGPLCD_WIDTH*3+1] = NGPVGA_CFG_BORDERBR ;
	p_buff[80+NGPLCD_WIDTH*3+2] = NGPVGA_CFG_BORDERBR ;

	drawGradationData(&m_ngpData[line][0][0], &p_buff[80], &m_convColor[m_color][0]) ;
}
