// -------------------------------------------------------------------------------- 
//!  @file  vga.c                                                                   
//!  VGA ソースファイル                                                             
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <string.h>
#include <pico/stdlib.h>
#include <pico/multicore.h>
#include <hardware/irq.h>
#include "config.h"
#include "vga.h"
#include "vga.pio.h"
#include "ngplcd.h"

#if     (VGA_CFG_PIO == 0)
#define PIO							pio0							//!< PIO オブジェクト
#define PIO_IRQ						PIO0_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO0_TX0					//!< PIO データリクエスト
#else
#define PIO							pio1							//!< PIO オブジェクト
#define PIO_IRQ						PIO1_IRQ_0						//!< PIO 割り込みリクエスト
#define PIO_DREQ					DREQ_PIO1_TX0					//!< PIO データリクエスト
#endif/*(VGA_CFG_PIO == 0)*/

#if     (VGA_CFG_DMA == 0)
#define DMA_IRQ						DMA_IRQ_0						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints0							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq0_enabled	//!< DMA 有効処理
#else
#define DMA_IRQ						DMA_IRQ_1						//!< DMA 割り込みリクエスト
#define DMA_INTS					ints1							//!< DMA 割り込みステータス
#define dma_enable					dma_channel_set_irq1_enabled	//!< DMA 有効処理
#endif//(VGA_CFG_DMA == 0)

	//! PIO ステートマシン番号
	enum {
		SM_CLKGEN_HORZ,												//!< VGA 水平クロックジェネレータ
		SM_CLKGEN_VERT,												//!< VGA 垂直クロックジェネレータ
		SM_TRANS_PIXEL												//!< VGA 有効ピクセル転送
	} ;

	//! ステートマシン情報
	typedef struct {
		uint8_t						pin ;							//!< ピン開始番号
		uint8_t						count ;							//!< ピン数
		uint8_t						sideset ;						//!< サイドセットフラグ
		uint32_t					out ;							//!< ピン出力値
		struct pio_program			const *p_program ;				//!< プログラム
		pio_sm_config (*getConfig)( uint offset ) ;					//!< コンフィグ取得処理
		uint32_t					scratchY ;						//!< スクラッチレジスター Y 初期値
	} PIOSMINFO ;


VGAINFO								g_vgaInfo ;						//!< VGA 情報

//! PIO ステートマシン情報
static const PIOSMINFO				m_pioSmInfo[]={
										{ VGA_CFG_PIN_HSYNC, 1,                      1, 1, &vga_horz_program,  &vga_horz_program_get_default_config,  (((48+640+16)*VGA_CFG_DOTCLK-1-1)<<16) | ( 96    *VGA_CFG_DOTCLK-3-1) },
										{ VGA_CFG_PIN_VSYNC, 1,                      1, 1, &vga_vert_program,  &vga_vert_program_get_default_config,  ((480                         -1)<<16) | ( 33-2                   -1) },
										{ VGA_CFG_PIN_PIXEL, VGA_CFG_PIN_PIXELCOUNT, 0, 0, &vga_pixel_program, &vga_pixel_program_get_default_config, ((VGA_IMAGE_WIDTH/2           -1)<<16) | ((96+48)*VGA_CFG_DOTCLK-4-1) }
									} ;


// -------------------------------------------------------------------------------- 
//!  PIO 割り込みサービスルーチン                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC isr_pio( void )
{
	dma_channel_abort( VGA_CFG_DMACH_PIXEL ) ;

	PIO->irq            = 1 << 1 ;
	g_vgaInfo.dispLine  = 0 ;
	g_vgaInfo.drawLine  = 1 ;
	g_vgaInfo.drawCount = 0 ;
}


// -------------------------------------------------------------------------------- 
//!  DMA 割り込みサービスルーチン                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC isr_dma( void )
{
	if( dma_hw->DMA_INTS & (1<<VGA_CFG_DMACH_PIXEL) ) {
		dma_hw->DMA_INTS = (1<<VGA_CFG_DMACH_PIXEL) ;

		if( ++g_vgaInfo.dispLine < VGA_HEIGHT ) {

			dma_channel_configure(VGA_CFG_DMACH_PIXEL, &g_vgaInfo.dmaConfigPixel, &PIO->txf[SM_TRANS_PIXEL], &g_vgaInfo.lineBuff.page[(g_vgaInfo.dispLine/3)&1].data[0], (VGA_IMAGE_WIDTH*2)/4, true) ;

			if( ++g_vgaInfo.drawCount == 2 ) {
				g_vgaInfo.drawCount = 0 ;
				g_vgaInfo.drawLine  = g_vgaInfo.dispLine / 3 + 1 ;
			}

		}
	}
}


// -------------------------------------------------------------------------------- 
//!  PIO ステートマシンを初期化                                                     
//!                                                                                 
//!  @param [in] sm           ステートマシンインデックス                            
//!  @param [in] p_pioSmInfo  ステートマシン情報アドレス                            
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
static void initPioSm(uint sm, const PIOSMINFO *p_pioSmInfo)
{
	pio_sm_config	config ;
	int				index ;
	uint			offset ;
	uint32_t		out ;

	// GPIO 初期化
	out = p_pioSmInfo->out ;
	for(index=0; index<p_pioSmInfo->count; index++,out>>=1) {
		pio_gpio_init(PIO, p_pioSmInfo->pin+index) ;
		gpio_disable_pulls( p_pioSmInfo->pin+index ) ;
		gpio_put(p_pioSmInfo->pin+index, out&1) ;
	}

	pio_sm_set_consecutive_pindirs(PIO, sm, p_pioSmInfo->pin, p_pioSmInfo->count, true) ;

	offset = pio_add_program(PIO, p_pioSmInfo->p_program) ;
	config = p_pioSmInfo->getConfig( offset ) ;

	sm_config_set_set_pins(&config, p_pioSmInfo->pin, p_pioSmInfo->count) ;
	sm_config_set_out_pins(&config, p_pioSmInfo->pin, p_pioSmInfo->count) ;

	if( p_pioSmInfo->sideset ) {
		sm_config_set_sideset_pins(&config, p_pioSmInfo->pin) ;
	}

	pio_sm_init(PIO, sm, offset, &config) ;

	// スクラッチレジスター Y 初期化
	pio_sm_put_blocking(PIO, sm, p_pioSmInfo->scratchY) ;
	pio_sm_exec_wait_blocking(PIO, sm, pio_encode_pull(false, false)) ;
	pio_sm_exec_wait_blocking(PIO, sm, pio_encode_mov(pio_y, pio_osr)) ;
}


// -------------------------------------------------------------------------------- 
//!  内部情報を初期化                                                               
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void init( void )
{
	// ステートマシン
	initPioSm(SM_CLKGEN_HORZ, &m_pioSmInfo[SM_CLKGEN_HORZ]) ;
	initPioSm(SM_CLKGEN_VERT, &m_pioSmInfo[SM_CLKGEN_VERT]) ;
	initPioSm(SM_TRANS_PIXEL, &m_pioSmInfo[SM_TRANS_PIXEL]) ;

	// 一部モニターで同期がふらつくためドライブストレングスを調整（現状が仕様を満たしていないと思われるので要調査）
	hw_write_masked(&padsbank0_hw->io[VGA_CFG_PIN_VSYNC], PADS_BANK0_GPIO0_DRIVE_VALUE_8MA<<PADS_BANK0_GPIO0_DRIVE_LSB, PADS_BANK0_GPIO0_DRIVE_BITS) ;
	hw_write_masked(&padsbank0_hw->io[VGA_CFG_PIN_HSYNC], PADS_BANK0_GPIO0_DRIVE_VALUE_8MA<<PADS_BANK0_GPIO0_DRIVE_LSB, PADS_BANK0_GPIO0_DRIVE_BITS) ;

	// ピクセルデータ転送開始割り込み設定
	hw_set_bits(&PIO->inte0, 1<<(8+1)) ;				// +1 == SM1
	PIO->irq = 1 << 1 ;

	irq_set_exclusive_handler(PIO_IRQ, isr_pio) ;
	irq_clear( PIO_IRQ ) ;
	irq_set_enabled(PIO_IRQ, true) ;

	// ピクセル転送用 DMA
	dma_channel_claim( VGA_CFG_DMACH_PIXEL ) ;
	g_vgaInfo.dmaConfigPixel = dma_channel_get_default_config( VGA_CFG_DMACH_PIXEL ) ;
	channel_config_set_dreq(&g_vgaInfo.dmaConfigPixel, PIO_DREQ+SM_TRANS_PIXEL) ;
	dma_enable(VGA_CFG_DMACH_PIXEL, true) ;

	irq_set_exclusive_handler(DMA_IRQ, isr_dma) ;
	irq_clear( DMA_IRQ ) ;
	irq_set_enabled(DMA_IRQ, true) ;

	// データ転送用 DMA
	dma_channel_claim( VGA_CFG_DMACH_DATA ) ;
	g_vgaInfo.dmaConfigData = dma_channel_get_default_config( VGA_CFG_DMACH_DATA ) ;
	channel_config_set_read_increment(&g_vgaInfo.dmaConfigData, false) ;
	channel_config_set_write_increment(&g_vgaInfo.dmaConfigData, true) ;

	// PIO 動作開始
	pio_set_sm_mask_enabled(PIO, (1<<SM_TRANS_PIXEL)|(1<<SM_CLKGEN_VERT)|(1<<SM_CLKGEN_HORZ), true) ;
}


// -------------------------------------------------------------------------------- 
//!  メイン処理                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void VGAFUNC mainProc( void )
{
	uint32_t	line ;


	init() ;

	while( multicore_fifo_rvalid() == false ) {
		line = g_vgaInfo.drawLine ;
		if( line ) {

			// 先頭行
			if( line == 1 ) {
				ngplcd_drawLine(&g_vgaInfo.lineBuff.page[0].data[0], 0) ;
				dma_channel_configure(VGA_CFG_DMACH_PIXEL, &g_vgaInfo.dmaConfigPixel, &PIO->txf[SM_TRANS_PIXEL], &g_vgaInfo.lineBuff.page[0].data[0], (VGA_IMAGE_WIDTH*2)/sizeof( uint32_t ), true) ;
			}

			ngplcd_drawLine(&g_vgaInfo.lineBuff.page[line&1].data[0], line) ;

			g_vgaInfo.drawLine = 0 ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void vga_init( void )
{
	memset(&g_vgaInfo, 0, sizeof( g_vgaInfo )) ;
}


// -------------------------------------------------------------------------------- 
//!  開始                                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void vga_start( void )
{
	multicore_launch_core1( mainProc ) ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 待ち                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void VGAFUNC vga_waitVSync( void )
{
	while( !gpio_get( VGA_CFG_PIN_VSYNC ) ) {
		;
	}

	while( gpio_get( VGA_CFG_PIN_VSYNC ) ) {
		asm volatile( "wfi" ) ;
	}
}
